# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Interactive command line interface to EGEGrouper."""

import cmd, sys, argparse, readline
from collections import OrderedDict

from . import sqlite3_model
from . import text_views
from . import plot_views
from . import controller

class DialogText:
    """Text dialog for input fields values. 

    This dialog processes data in the dictionary with keys which correspond attributes of group.

    """
    def __init__(self, data_dict):
        """Constructor.

        Parameters
        ----------
        data_dict : OrderedDict
            Dictionary to process.

        """
        self.data_dict = data_dict

    def input(self):
        """Process input data."""
        for key in self.data_dict:
            self.data_dict[key] = self.rlineinput(key + ': ', self.data_dict[key])
            
    def rlineinput(self, prompt, prefill=''):
        """Output existing value to the edit area."""
        readline.set_startup_hook(lambda: readline.insert_text(prefill))
        try:
            return input(prompt)
        finally:
            readline.set_startup_hook()
            

class GrouperShell(cmd.Cmd):
    """Class for processing user's commands."""
    
    prompt = 'igrouper> '

    def __init__(self):
        """Constructor.

        Set aliases for commands.
        """
        cmd.Cmd.__init__(self)
        self.aliases = {
            'd' : self.do_db_info,
            'g' : self.do_group_info,
            'e' : self.do_exam_info,
            'p' : self.do_plot_exam,
            
            'ag' : self.do_add_group,
            'dg' : self.do_delete_group,
            'at' : self.do_add_to_group,
            'df' : self.do_delete_from_group,
            'we' : self.do_where_is,
            
            'aj' : self.do_add_json,
            'ej' : self.do_export_json
            }

    def default(self, line):
        """This method called if user gives not existing command.

        Here the aliases processed.

        Parameters
        ----------
        line : str
            User's command.

        """
        cmd, arg, line = self.parseline(line)
        if cmd in self.aliases:
            self.aliases[cmd](arg)
        else:
            print("*** Unknown syntax: %s" % line)

    def parse(self, arg, option_name, has_value = False):
        args = arg.split()
        for (a, i) in zip(args, range(0, len(args))):
            if (a == option_name):
                if not has_value:
                    return True
                else:
                    if len(args) > i+1:
                        return args[i+1]
        return None

    def do_EOF(self, arg):
        print()
        return True

    def do_quit(self, arg):
        """ Close data base and exit.
        """
        controller.close_storage()
        return True

    def do_db_info(self, arg):
        """ Print information about data base.

        Aliases: d
        """
        controller.storage_info()

    def do_group_info(self, arg):
        """ group_info id
        
        Print information about group.

        Aliases: g
        """
        cargv = arg.split()
        if len(cargv) == 0:
            print('Few arguments')
            return
        controller.group_info(cargv[0])

    def do_exam_info(self, arg):
        """ exam_info id
        
        Print information about examination.

        Aliases: e
        """
        cargv = arg.split()
        if len(cargv) < 1:
            print('Few arguments')
            return
        controller.exam(cargv[0])

    def do_plot_exam(self, arg):
        """ plot_exam id
        
        Plot signals of examination.

        Aliases: p
        """
        cargv = arg.split()
        if len(cargv) < 1:
            print('Few arguments')
            return
        controller.plot_exam(cargv[0])
            
    def do_add_group(self, arg):
        """ add_group

        Add new group.

        Aliases: ag
        """
        data = OrderedDict([('Name', ''), ('Description', '')])
        DialogText(data).input()
        controller.insert_group(data['Name'], data['Description'])

    def do_delete_group(self, arg):
        """ delete_group id

        Delete group

        Aliases: dg
        """
        cargv = arg.split()
        if len(cargv)==0:
            print('Few arguments')
            return
        answer = input('Are you sure? (yes/no/y/n): ')
        if answer not in ['yes', 'y']:
            return
        controller.delete_group(cargv[0])
        
    def do_add_to_group(self, arg):
        """ add_to_group exam_id group_id

        Add examination to group.

        Aliases: at
        """
        cargv = arg.split()
        if len(cargv) < 2:
            print('Few arguments')
            return
        controller.group_exam(cargv[0], [cargv[1], ], [True, ])

    def do_delete_from_group(self, arg):
        """ delete_from_group exam_id group_id

        Delete examination from group.

        Aliases: df
        """
        cargv = arg.split()
        if len(cargv) < 2:
            print('Few arguments')
            return
        controller.group_exam(cargv[0], [cargv[1], ], [False, ])

    def do_where_is(self, arg):
        """ where_is id

        Show where is examination.

        Aliases: we
        """
        cargv = arg.split()
        if len(cargv) == 0:
            print('Few arguments')
            return
        controller.where_exam(cargv[0])

    def do_add_sme(self, arg):
        """ add_sme file_name

        Add records from SME data base.
        """
        cargv = arg.split()
        if len(cargv) == 0:
            print('Few arguments')
            return
        controller.add_sme_db(cargv[0])

    def do_add_gs(self, arg):
        """ add_gs file_name

        Add records from Gastroscan sqlite data base.
        """
        cargv = arg.split()
        controller.add_gs_db(cargv[0])

    def do_add_json(self, arg):
        """ add_json file_name

        Add examination from JSON file.

        Aliases: aj
        """
        cargv = arg.split()
        if len(cargv) == 0:
            print('Few arguments')
            return
        controller.add_exam_from_json_file(cargv[0])

    def do_export_json(self, arg):
        """ export_json exam_id file_name

        Export examination to JSON file.

        Aliases: ej
        """
        cargv = arg.split()
        if len(cargv) < 2:
            print('Few arguments')
            return
        exam_id = cargv[0]
        folder_name = cargv[1]
        controller.export_exam_to_json_file(exam_id, folder_name)

    def do_delete_exam(self, arg):
        """ delete_exam id

        Delete examination from data base.

        Aliases: de
        """
        cargv = arg.split()
        if len(cargv) == 0:
            print('Few arguments')
            return
        
        answer = input('Are you sure? (yes/no/y/n): ')
        if answer not in ['yes', 'y']:
            return
        
        exam_id = cargv[0]
        controller.delete_exam(exam_id)

    def do_merge_exams(self, arg):
        """ merge_exams exam_id_1 exam_id_2

        Merge two examinations."""
        cargv = arg.split()
        if len(cargv) < 2:
            print('Few arguments')
            return
        exam_id_1 = cargv[0]
        exam_id_2 = cargv[1]
        controller.merge_exams(exam_id_1, exam_id_2)

    def do_edit_group(self, arg):
        """edit_group group_id

        Edit attributes of selected group.
        
        """
        cargv = arg.split()
        if len(cargv) < 1:
            print('Few arguments')
            return
        group_id = cargv[0]
        
        data = controller.group_record(group_id)
        if not data:
            return
        DialogText(data).input()
        controller.update_group_record(group_id, data)

    def do_c(self, arg):
        print("""
        This program is free software: you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation, either version 3 of the License, or
        (at your option) any later version.
        """)

    def do_w(self, arg):
        print("""
        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.
        """)    

controller = controller.Controller(sqlite3_model.Model())
"""Controller."""
        
def main():
    """Entry point."""
    
    print("""
    EGEGrouper 0.3.1 Copyright (C) 2017 Aleksandr Popov

    This program comes with ABSOLUTELY NO WARRANTY; for details type `w'.
    This is free software, and you are welcome to redistribute it
    under certain conditions; type `c' for details.
    """)

    parser = argparse.ArgumentParser()
    parser.add_argument("fname", help="Name of data base")
    args = parser.parse_args()
    
    controller.set_view_message(text_views.Message())
    controller.set_view_storage(text_views.Table())
    controller.set_view_group(text_views.Table())
    controller.set_view_exam(text_views.Exam())
    controller.set_view_where_exam(text_views.WhereExam())
    controller.set_view_exam_plot(plot_views.Exam())
    
    if not controller.open_or_create_storage(args.fname):
        print("Can't open storage\nExit...\n")
        return

    gshell = GrouperShell()
    gshell.cmdloop()



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Mapping examinations to and from JSON files."""

import json
import numpy as np
import os, shutil
from collections import OrderedDict

from . import sme

def get_exam(file_name):
    """Get examination object from JSON file.

    Parameters
    ----------
    file_name : str
        Name of JSON file.
    
    Return
    ------
    : sme.Examination
        Examination instance.

    """
    abs_file_name = os.path.expanduser(file_name)
    abs_data_folder_name = "{}.data".format(abs_file_name)
    e = sme.Examination()
    with open(abs_file_name, 'r') as f:
        data = json.load(f)
    e.name = data['name']
    e.age = data['age']
    e.gender = data['gender']
    e.diagnosis = data['diagnosis']
    e.ms = []
    for m_data in data['measurements']:
        m = sme.Measurement()
        m.time = m_data['time']
        m.ss = []
        for s_data in m_data['signals']:
            s = sme.Signal()
            s.dt = s_data['dt']
            s.x = np.loadtxt(
                os.path.join(
                    abs_data_folder_name,
                    s_data['file']
                )
            )
            m.ss.append(s)
        e.ms.append(m)
    return e

def put_exam(e, file_name):
    """Put examination to JSON file.

    Parameters
    ----------
    e : sme.Examination
        Examination instance.
    file_name : str
        File name.
    
    """
    abs_file_name = os.path.expanduser(file_name)
    abs_data_folder_name = "{}.data".format(abs_file_name)
    if os.path.isfile(abs_file_name):
        os.remove(abs_file_name)
    if os.path.isdir(abs_data_folder_name):
        shutil.rmtree(abs_data_folder_name)
    os.makedirs(abs_data_folder_name)
    e_dict = OrderedDict()
    e_dict['name'] = e.name
    e_dict['diagnosis'] = e.diagnosis
    e_dict['age'] = e.age
    e_dict['gender'] = e.gender
    e_dict['measurements'] = []
    for (m, mn) in zip(e.ms, range(1, len(e.ms)+1)):
        m_dict = OrderedDict()
        m_dict['time'] = m.time
        m_dict['signals'] = []
        for (s, sn) in zip(m.ss, range(1, len(m.ss)+1)):
            s_dict = OrderedDict()
            s_dict['dt'] = s.dt
            signal_file_name = 'signal-{}{}.txt'.format(mn, sn)
            s_dict['file'] = signal_file_name
            np.savetxt(
                os.path.join(
                    abs_data_folder_name,
                    signal_file_name
                ),
                s.x,
                fmt='%f'
            )
            m_dict['signals'].append(s_dict)
        e_dict['measurements'].append(m_dict)
    s = json.dumps(e_dict, ensure_ascii=False, indent='    ')
    with open(abs_file_name, 'w') as f:
        f.write(s)



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""View for plotting data here."""

import matplotlib.pyplot as plt

class Exam:
    """Plot view to show details of examination."""
    
    def show_data(self, exam):
        """Plot signals of examination with matplotlib.

        Parameters
        ----------
        exam: sme.Examination
            SME examination object.

        """
        plt.ion()
        plt.clf()
        n = 0
        for m in exam.ms:
            for s in m.ss:
                n = n + 1
        i = 0
        for m in exam.ms:
            t = m.time
            for s in m.ss:
                i = i + 1
                plt.subplot(n*100 + 10 + i)
                plt.plot(s.x, 'b', linewidth=1)
                plt.xlim(0,len(s.x))
                plt.grid(True)
                plt.title(t)

        plt.tight_layout()
        plt.show()



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tk interface to egegrouper."""

from tkinter import *
from tkinter import filedialog
from tkinter import messagebox

from . import controller
from . import sqlite3_model
from . import tk_views
from . import plot_views

from collections import OrderedDict

class MainWindow:
    """Main window. Shows groups and main menu."""
    
    def __init__(self):
        """Constructor.

        Create main window.
        """
        self.master = Tk()
        self.master.title("EGEGrouper 0.3.1")

        # menu
        self.main_menu = Menu(self.master)
        self.storage_menu = Menu(self.main_menu, tearoff=0)
        self.storage_menu.add_command(label="Open", command=self.open_storage)
        self.storage_menu.add_command(label="Create", command=self.create_storage)
        self.storage_menu.add_command(label="Close", command=self.close_storage)
        self.add_data_submenu = Menu(self.storage_menu, tearoff=0)
        self.add_data_submenu.add_command(label="Add SME sqlite3 DB", command=self.add_sme)
        self.add_data_submenu.add_command(label="Add exam from JSON", command=self.add_json)
        #self.add_data_submenu.add_command(label="TODO: Add Gastroscan sqlite3 DB", command=)
        self.storage_menu.add_cascade(label="Add data", menu=self.add_data_submenu)
        self.storage_menu.add_command(label="Exit", command=self.close_db_and_exit)
        self.main_menu.add_cascade(label="Storage", menu=self.storage_menu)
        self.group_menu = Menu(self.main_menu, tearoff=0)
        self.group_menu.add_command(label="Add", command=self.add_group)
        self.group_menu.add_command(label="Edit", command=self.edit_group)
        self.group_menu.add_command(label="Delete", command=self.delete_group)
        self.main_menu.add_cascade(label="Group", menu=self.group_menu)
        self.exam_menu = Menu(self.main_menu, tearoff=0)
        self.exam_menu.add_command(label="Plot", command=self.plot_exam)
        self.exam_menu.add_command(label="Grouping", command=self.grouping)
        self.exam_menu.add_command(label="Delete forever", command=self.delete_exam)
        #self.exam_menu.add_command(label="TODO?:Merge with", command=None)
        self.exam_menu.add_command(label="Export to JSON", command=self.export_json)
        self.main_menu.add_cascade(label="Exam", menu=self.exam_menu)
        self.help_menu = Menu(self.main_menu, tearoff=0)
        self.help_menu.add_command(label="About", command=self.about)
        self.main_menu.add_cascade(label="Help", menu=self.help_menu)
        self.master.config(menu=self.main_menu)

        self.storage_menu.entryconfig("Add data", state=DISABLED)
        self.storage_menu.entryconfig("Close", state=DISABLED)
        self.main_menu.entryconfig("Group", state=DISABLED)
        self.main_menu.entryconfig("Exam", state=DISABLED)

        self.view_storage = tk_views.Storage(self.master)
        self.view_storage.pack(side=LEFT, fill=BOTH, expand=True)
        self.view_storage.item_opened.connect(self.group_info)
        self.view_storage.item_selected.connect(self.group_selected)
        controller.set_view_storage(self.view_storage)
        self.group_window = GroupWindow(self.master)
        self.view_group = self.group_window.view_group
        self.view_group.item_opened.connect(self.plot_exam)
        self.view_group.item_selected.connect(self.exam_selected)
        controller.set_view_group(self.view_group)
        controller.set_view_exam_plot(plot_views.Exam())

    def open_storage(self):
        """Open storage and show groups in it."""
        file_name = filedialog.askopenfilename(
            title='Open storage',
            filetypes = [('sme db files', '.sme.sqlite'), ('all files', '.*')],
            parent = self.master,
        )
        if not file_name:
            return
        controller.open_storage(file_name)
        self.view_group.clear()
        # menu
        self.storage_menu.entryconfig("Add data", state=NORMAL)
        self.storage_menu.entryconfig("Close", state=NORMAL)
        self.main_menu.entryconfig("Group", state=NORMAL)
        self.group_menu.entryconfig("Edit", state=DISABLED)
        self.group_menu.entryconfig("Delete", state=DISABLED)
        self.main_menu.entryconfig("Exam", state=DISABLED)
        
    def create_storage(self):
        """Create storage."""
        file_name = filedialog.asksaveasfilename(
            title='Open storage',
            defaultextension = '.sme.sqlite',
            filetypes = [('sme db files', '.sme.sqlite'), ('all files', '.*')],
            parent = self.master,
        )
        if not file_name:
            return
        controller.create_storage(file_name)
        self.view_group.clear()
        # menu
        self.storage_menu.entryconfig("Add data", state=NORMAL)
        self.storage_menu.entryconfig("Close", state=NORMAL)
        self.main_menu.entryconfig("Group", state=NORMAL)
        self.group_menu.entryconfig("Edit", state=DISABLED)
        self.group_menu.entryconfig("Delete", state=DISABLED)
        self.main_menu.entryconfig("Exam", state=DISABLED)

    def close_storage(self):
        """Close storage and clear widgets."""
        controller.close_storage()
        self.view_group.clear()
        self.view_storage.clear()
        # menu
        self.storage_menu.entryconfig("Add data", state=DISABLED)
        self.storage_menu.entryconfig("Close", state=DISABLED)
        self.main_menu.entryconfig("Group", state=DISABLED)
        self.main_menu.entryconfig("Exam", state=DISABLED)

    def add_sme(self):
        "Add records from sqlite3 data base in SME format."
        file_name = filedialog.askopenfilename(
            title='Open storage',
            filetypes = [('sme db files', '.sme.sqlite'), ('all files', '.*')],
            parent = self.master,
        )
        if not file_name:
            return
        controller.add_sme_db(file_name)
        controller.storage_info()

    def add_json(self):
        """Add examination from JSON file."""
        file_name = filedialog.askopenfilename(
            parent = self.master,
            title = 'Add exam from JSON',
            filetypes = [('JSON files', '.json'), ('all files', '.*')],
        )
        if not file_name:
            return
        controller.add_exam_from_json_file(file_name)
        controller.storage_info()

    def group_selected(self, *args):
        """Group selected slot. Enable some menu items."""
        # menu
        self.group_menu.entryconfig("Edit", state=NORMAL)
        self.group_menu.entryconfig("Delete", state=NORMAL)

    def exam_selected(self, *args):
        """Exam selected slot. Enable some menu items."""
        # menu
        self.main_menu.entryconfig("Exam", state=NORMAL)

    def group_info(self, *args):
        """Get and show information about examination in selected group."""
        if self.group_window.master.state() == "withdrawn":
            self.group_window.master.deiconify()
        group_id = self.view_storage.selected_item_text()
        if group_id:
            controller.group_info(group_id)
            self.view_storage.last_group_id = group_id
        # menu
        self.main_menu.entryconfig("Exam", state=DISABLED)

    def grouping(self):
        """Open grouping dialog stub."""
        exam_id = self.view_group.selected_item_text()
        if not exam_id:
            return

        self.view_storage.remember_selection()
        self.view_group.remember_selection()
        grouping_dialog = GroupingDialog(self.master, exam_id)
        grouping_dialog.master.transient(self.master)
        grouping_dialog.master.grab_set()
        grouping_dialog.master.wait_window(grouping_dialog.master)
        controller.storage_info()
        controller.group_info(self.view_storage.last_group_id)
        self.view_storage.restore_selection()
        self.view_group.restore_selection()
        
    def delete_exam(self):
        """Delete exam from storage."""
        exam_id = self.view_group.selected_item_text()
        if not exam_id:
            return
        if messagebox.askquestion("Delete examination", "Are You shure?", icon='warning') == 'no':
            return
        controller.delete_exam(exam_id)
        self.view_storage.remember_selection()
        self.view_group.remember_selection()
        controller.storage_info()
        controller.group_info(self.view_storage.last_group_id)
        self.view_storage.restore_selection()
        self.view_group.restore_selection()

    def export_json(self):
        """Export selected examination to JSON file."""
        exam_id = self.view_group.selected_item_text()
        if not exam_id:
            return
        file_name = filedialog.asksaveasfilename(
            parent = self.master,
            title = 'Export exam to JSON',
            defaultextension = '.json',
            filetypes = [('JSON files', '.json'), ('all files', '.*')],
        )
        if not file_name:
            return
        controller.export_exam_to_json_file(exam_id, file_name)
        
    def add_group(self):
        """Add new group."""
        group_record_dialog = GroupRecordDialog(
            self.master,
            group_record = OrderedDict([('name',''), ('description','')])
        )
        group_record_dialog.master.transient(self.master)
        group_record_dialog.master.grab_set()
        group_record_dialog.master.wait_window(group_record_dialog.master)

    def edit_group(self):
        """Edit group."""
        group_id = self.view_storage.selected_item_text()
        if not group_id:
            return
        if group_id=='0':
            return
        self.view_storage.remember_selection()
        data = controller.group_record(group_id)
        group_record_dialog = GroupRecordDialog(self.master, data, group_id)
        group_record_dialog.master.transient(self.master)
        group_record_dialog.master.grab_set()
        group_record_dialog.master.wait_window(group_record_dialog.master)
        self.view_storage.restore_selection()

    def delete_group(self):
        """Delete selected group."""
        group_id = self.view_storage.selected_item_text()
        if not group_id:
            return
        if group_id=='0':
            return
        if messagebox.askquestion("Delete", "Are You shure?", icon='warning') == 'no':
            return
        self.view_storage.remember_selection()
        controller.delete_group(group_id)
        controller.storage_info()
        self.view_storage.restore_selection()

    def plot_exam(self, *args):
        """Plot examination in separate matplotlib window."""
        exam_id = self.view_group.selected_item_text()
        if exam_id:
            controller.plot_exam(exam_id)
                
    def close_db_and_exit(self):
        """Close data base and exit."""
        controller.close_storage()
        self.master.quit()

    def about(self):
        """Show info about program."""
        about_window = AboutWindow(self.master)
        about_window.master.transient(self.master)
        about_window.master.grab_set()
        about_window.master.wait_window(about_window.master)

class GroupWindow:
    """Window for show and select examinations."""
    
    def __init__(self, parent):
        """Constructor.

        Create window.

        Parameters
        ----------
        parent
            Master for window.

        """
        self.master = Toplevel(parent)
        self.master.title("Examinations")
        self.master.protocol('WM_DELETE_WINDOW', self.on_destroy)
        self.view_group = tk_views.Group(self.master)
        self.view_group.pack(side=LEFT, fill=BOTH, expand=True)

    def on_destroy(self):
        """Do not destroy, but withdraw."""
        self.master.withdraw()

class GroupingDialog:
    """Dialog for grouping examinations."""
    
    def __init__(self, parent, exam_id):
        """Constructor.
        
        Parameters
        ----------
        parent
            Master.
        exam_id : str
            Examination ID.

        """
        self.exam_id = exam_id
        self.master = Toplevel(parent)
        self.master.title("Grouping")
        self.grouping_widget = tk_views.WhereExam(self.master)
        self.save_button = Button(self.master, text="Save", width=15, command=self.on_save_button)
        self.cancel_button = Button(self.master, text="Cancel", width=15, command=self.master.destroy)
        self.grouping_widget.pack()
        self.cancel_button.pack(side=RIGHT)
        self.save_button.pack(side=RIGHT)
        controller.set_view_where_exam(self.grouping_widget)
        controller.where_exam(exam_id)

    def on_save_button(self):
        """Save button handler."""
        group_ids, placed_in = self.grouping_widget.checked_group_ids()
        controller.group_exam(self.exam_id, group_ids, placed_in)
        self.master.destroy()

class GroupRecordDialog:
    """Dialog for edit group record."""
    
    def __init__(self, parent, group_record, group_id = None):
        """Constructor.
        
        Parameters
        ----------
        parent
            Master.
        group_record : OrderedDict
            Attributes and values of group.
        group_is : str
            Group ID.

        """
        self.group_id = group_id
        self.group_record = group_record
        self.master = Toplevel(parent)
        self.master.title("Edit group")
        self.labels = []
        self.entries = []
        for key in group_record:
            label = Label(self.master, width = 10, text=key)
            label.pack(side=TOP)
            entry = Entry(self.master, width = 30)
            entry.delete(0, END)
            if group_record[key]:
                entry.insert(0, group_record[key])
            else:
                entry.insert(0, '')
            entry.pack(side=TOP)
            self.labels.append(label)
            self.entries.append(entry)
        self.save_button = Button(self.master, text="Save", width=15, command=self.on_save_button)
        self.cancel_button = Button(self.master, text="Cancel", width=15, command=self.master.destroy)
        self.cancel_button.pack(side=RIGHT)
        self.save_button.pack(side=RIGHT)

    def on_save_button(self):
        """Save button handler."""
        for (key, e) in zip(self.group_record, self.entries):
            self.group_record[key] = e.get()
        if not self.group_id:
            controller.insert_group(self.group_record['name'], self.group_record['description'])
        else:
            controller.update_group_record(self.group_id, self.group_record)
        controller.storage_info()
        self.master.destroy()

class AboutWindow:
    """Window with short info about the program."""
    def __init__(self, parent):
        """Constructor.

        Create window with info about the program.

        """
        self.master = Toplevel(parent)
        self.master.title("About EGEGrouper")
        label = Label(self.master, text="""
        EGEGrouper Copyright (C) 2017 Aleksandr Popov

        This program comes with ABSOLUTELY NO WARRANTY.
        This is free software, and you are welcome to redistribute it
        under certain conditions.
        """)
        label.pack(side=TOP)
        self.close_button = Button(self.master, text="Close", width=15, command=self.master.destroy)
        self.close_button.pack(side=TOP)

controller = controller.Controller(sqlite3_model.Model())
"""Controller."""
controller.set_view_message(tk_views.Message())

def main():
    """Entry point."""
    main_window = MainWindow()
    main_window.master.mainloop()



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod

class BaseModel(ABC):
    """Model in MVC. Base class."""
    
    def __init__(self):
        """Constructor.
        
        Create fields and set initial state of model.
        
        """
        self._state = {}

    def set_state(self, **kwargs):
        """Set model state."""
        for key in kwargs:
            self._state[key] = kwargs[key]

    def state(self):
        """Return model state."""
        return self._state

    @staticmethod
    def do_if_storage_opened(method):
        """Decorator. If storage is not opened AttributeError raised."""
        def wrapped(self, *args):
            if not self.state()['storage_opened']:
                raise AttributeError('Storage is not opened.')
            return method(self, *args)
        wrapped.__doc__ = method.__doc__
        return wrapped

    def open_or_create_storage(self, name):
        """Open or create storage (if stoarge not exists).
        
        Parameters
        ----------
        name : str
            Name of storage.
        
        """
        if self.storage_exists(name):
            self.open_storage(name)
        else:
            self.create_storage(name)

    @abstractmethod
    def create_storage(self, name):
        """Create new storage.

        Parameters
        ----------
        name : str
            Storage name.

        """
        pass

    @abstractmethod
    def open_storage(self, name):
        """Open storage.

        Parameters
        ----------
        name : str
            Storage name.

        """
        pass

    @abstractmethod
    def close_storage(self):
        """Close current storage."""
        pass

    @abstractmethod
    def storage_exists(self, name):
        """Check if the storage exists.

        Parameters
        ----------
        name : str
            Name of storage.

        Returns
        -------
        : bool
            True if exists, False otherwise.

        """
        pass

    @abstractmethod
    def exam(self, exam_id):
        """Return examination object.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        Returns
        -------
        sme.Examination
            Examination object.

        """
        pass

    @abstractmethod
    def insert_exam(self, exam):
        """Add examination into current storage.

        Parameters
        ----------
        exam : sme.Examination
            Examination object

        """
        pass

    @abstractmethod
    def delete_exam(self, exam_id):
        """Delete examination from current storage.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        """
        pass
    
    @abstractmethod
    def storage_info(self):
        """Return common information about current storage.

        Return
        ------
        data : list of tuples
            Table with information about storage.
        headers : tuple
            Headers.

        """
        pass

    @abstractmethod
    def group_info(self, group_id):
        """Return information about examinations of selected group.

        Parameters
        ----------
        group_id : str
            Group ID

        Returns
        -------
        data : list of tuple
            Examination descriptions.
        headers : tuple
            Headers.

        """
        pass

    @abstractmethod
    def exams(self, group_id):
        """
        Return examinations of selected group.

        Parameters
        ----------
        group_id : str
           Group ID.

        Returns
        -------
        : list of sme.Examination
            Examination objects.
        
        """
        pass
    
    @abstractmethod
    def insert_group(self, name, description):
        """Add new group of examinations.

        Parameters
        ----------
        name : str
            Name of new group.
        description : str
            Description for new group.

        """
        pass

    @abstractmethod
    def delete_group(self, group_id):
        """Delete group of examinations from storage.

        Parameters
        ----------
        group_id : str
            Group ID.

        """
        pass

    @abstractmethod
    def group_exam(self, exam_id, group_ids, placed_in):
        """Add and delete examination to and from groups.

        Parameters
        ----------
        exam_id : str
            Examination ID.
        group_ids : list of str
            Group IDs.
        placed_in : list of bool
            True for examinations to be placed in groups. Length of group_ids must be equal to length of placed_in.

        """
        pass

    @abstractmethod
    def where_exam(self, exam_id):
        """Return description of groups where examination in or not in.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        Returns
        -------
        group_records : list of tuple
            All group records.
        headers : list of str
            Names of group attributes.
        placed_in : list of bool
            True if exam in group, False otherwise.
        
        """
        pass

    @abstractmethod
    def group_record(self, group_id):
        """Return attribute names and values of selected group.

        Parameters
        ----------
        group_id : str
            Group ID.

        Returns
        -------
        : OrderedDict
            Attributes names and values for selected group.

        """
        pass

    @abstractmethod
    def update_group_record(self, group_id, attr):
        """Update group record in storage.

        Parameters
        ----------
        group_id : str
            Group ID.
        attr : OrderedDict
            Attributes names and values.

        """
        pass



# EGEGrouper - Software for grouping electrogastroenterography examinations.

# Copyright (C) 2017 Aleksandr Popov

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the Controller class.

The common scenario for controller is to ask connected model to make some data manipulation and then to ask suitable view to show data.

"""

from . import sme

class Controller:
    """Controller in MVC."""    

    def __init__(self, model):
        """Constructor. Set model.

        Parameters
        ----------
        model : egegrouper.GrouperModel
            Model.

        """
        self._model = model
        self._view_message = None
        self._view_storage = None
        self._view_group = None
        self._view_exam = None
        self._view_exam_plot = None
        self._view_where_exam = None

    def set_view_message(self, view):
        """Set view for show messages.

        Parameters
        ----------
        view
            View object.

        """
        self._view_message = view

    def set_view_storage(self, view):
        """Set view for show information about storage: groups, number of examinations in groups and number of ungrouped examination.

        Parameters
        ----------
        view
            View object.

        """
        self._view_storage = view
        
    def set_view_group(self, view):
        """Set view for show information about examinations in group.

        Parameters
        ----------
        view
            View object.

        """
        self._view_group = view
        
    def set_view_exam(self, view):
        """Set view for show information examination.

        Parameters
        ----------
        view
            View object.

        """
        self._view_exam = view
        
    def set_view_exam_plot(self, view):
        """Set view for plot signals of examination.

        Parameters
        ----------
        view
            View object.

        """
        self._view_exam_plot = view
        
    def set_view_where_exam(self, view):
        """Set view for show groups in which examination is.

        Parameters
        ----------
        view
            View object.

        """
        self._view_where_exam = view
        
    def show_message(self, text):
        """If controller has message view it asks view to show message.

        Parameters
        ----------
        text : str
            Text of message.

        """
        if not self._view_message:
            return
        self._view_message.show_data(text)

    def model_can_grumble(method):
        """Decorator. Try to do something and if model raises an exception return None.
        
        Parameters
        ----------
        method
            Method to decorate.

        """
        def wrapped(self, *args):
            try:
                return method(self, *args)
            except Exception as err:
                self.show_message('Something wrong') #TODO: more concrete messages
                #print(err.args) # uncomment for debugging
                return None
        wrapped.__doc__ = method.__doc__
        return wrapped

    @model_can_grumble
    def open_storage(self, file_name):
        """Open storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        file_name : str
            File name.

        """
        self._model.open_storage(file_name)
        self.storage_info()
        return True
        
    @model_can_grumble
    def create_storage(self, file_name):
        """Create storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        file_name : str
            File name.

        """
        self._model.create_storage(file_name)
        self.storage_info()
        return True

    @model_can_grumble
    def open_or_create_storage(self, file_name):
        """Open or create storage. Return True if success, None if an exception raised.

        If there are no data base it will be created.

        Parameters
        ----------
        file_name : str
            File name.

        """
        self._model.open_or_create_storage(file_name)
        self.storage_info()
        return True

    @model_can_grumble
    def close_storage(self):
        """Close storage. Return True if success, None if an exception raised."""
        self._model.close_storage()

    @model_can_grumble
    def storage_info(self):
        """Show common information about storage. Return True if success, None if an exception raised."""
        data, headers = self._model.storage_info()
        self._view_storage.show_data(data, headers)
        return True

    @model_can_grumble
    def group_info(self, group_id):
        """Show list of examinations of group. Return True if success, None if an exception raised.

        Parameters
        ----------
        group_id : str
            Group ID.

        """
        data, headers = self._model.group_info(group_id)
        self._view_group.show_data(data, headers)
        return True

    @model_can_grumble
    def exam(self, exam_id):
        """Show information about selected examination. Return True if success, None if an exception raised.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        """
        e = self._model.exam(exam_id)
        if e == None:
            self.show_message('Exam data is empty.')
            return True
        self._view_exam.show_data(e)
        return True

    @model_can_grumble
    def plot_exam(self, exam_id):
        """Plot signals of examination. Return True if success, None if an exception raised.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        """
        e = self._model.exam(exam_id)
        if e == None:
            self.show_message('Exam data is empty.')
            return True
        self._view_exam_plot.show_data(e)
        return True

    @model_can_grumble
    def insert_group(self, name, description):
        """Add new group to current storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        name : str
            Name of new group.
        description : str
            Description of new group.

        """
        self._model.insert_group(name, description)
        self.storage_info()
        return True

    @model_can_grumble
    def delete_group(self, group_id):
        """Delete group. Return True if success, None if an exception raised.

        Parameters
        ----------
        group_id : str
            Group ID.

        """
        self._model.delete_group(group_id)
        self.storage_info()
        return True

    @model_can_grumble
    def group_exam(self, exam_id, group_ids, placed_in):
        """Add and delete examination to and from groups. Return True if success, None if an exception raised.

        Parameters
        ----------
        exam_id : str
            Examination identifier.
        group_ids : list of str
            Group identifiers.
        placed_in : list of bool
            True for examinations to be placed in groups. Length of group_ids must be equal to length of placed_in.

        """
        self._model.group_exam(exam_id, group_ids, placed_in)
        return True

    @model_can_grumble
    def where_exam(self, exam_id):
        """Show information of groups where examination is. Return True if success, None if an exception raised.

        Parameters
        ----------
        exam_id : str
            Examination ID.

        """
        group_records, headers, placed_in = self._model.where_exam(exam_id)
        self._view_where_exam.show_data(group_records, headers, placed_in)
        return True

    @model_can_grumble
    def add_sme_db(self, file_name):
        """Add SME sqlite3 data base to current storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        file_name : str
            File name. Example: example.sme.sqlite.

        """
        self._model.add_sme_db(file_name)
        self.show_message('Done.')
        return True

    @model_can_grumble
    def add_gs_db(self, file_name):
        """Add GS sqlite3 data base to current storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        file_name : str
            File name. Example: example.gs.sqlite.

        """
        self._model.add_gs_db(file_name)
        self.show_message('Done.')
        return True

    @model_can_grumble
    def add_exam_from_json_file(self, file_name):
        """Add examination from JSON file to current storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        file_name : str
            Name of JSON file.
        
        """
        self._model.add_exam_from_json_file(file_name)
        self.show_message('Done.')
        return True

    @model_can_grumble
    def export_exam_to_json_file(self, exam_id, file_name):
        """Export examination to JSON file. Return True if success, None if an exception raised.

        Parameters
        ----------
        exam_id : str
            Examination ID.
        file_name : str
            Name of JSON file.

        """
        self._model.export_exam_to_json_file(exam_id, file_name)
        self.show_message('Done.')
        return True

    @model_can_grumble
    def delete_exam(self, exam_id):
        """Delete examination from storage. Return True if success, None if an exception raised.

        Parameters
        ----------
        exam_id : str
            ID of examination to be deleted.
        
        """
        self._model.delete_exam(exam_id)
        return True

    @model_can_grumble
    def merge_exams(self, exam_id_1, exam_id_2):
        """Merge two exams. Return True if success, None if an exception raised.
        
        Create joined examination and put it to storage. Add measurements from examination 2 to examination 1.

        Parameters
        ----------
        exam_id_1 : str
            Examination 1 ID.
        exam_id_2 : str
            Examination 2 ID.

        """
        e1 = self._model.exam(exam_id_1)
        e2 = self._model.exam(exam_id_2)
        e = sme.merge_exams(e1, e2)
        self._model.insert_exam(e)
        self.show_message('Done')
        return True

    @model_can_grumble
    def group_record(self, group_id):
        """Return group record.

        Parameters
        ----------
        group_id : str
            Group ID.

        Return
        ------
        : OrderedDict
            Attributes names and values. If an exception raised return None.

        """
        return self._model.group_record(group_id)

    @model_can_grumble
    def update_group_record(self, group_id, attr):
        """ Update attribute values of selected group. Return True if success, None if an exception raised.

        Parameters
        ----------
        group_id : str
            Group ID.
        attr : OrderedDict
            Attributes names and values.

        """
        self._model.update_group_record(group_id, attr)
        return True
